@@ -1,5 +1,6 @@
 # coding: utf-8
 import json
+import statistics
 from collections import Counter
 from functools import lru_cache
 from os.path import dirname, realpath, exists
@@ -85,7 +86,7 @@ class ProbeCoherence:
 
         ratios = [self.rank_per_lang[lg] for lg in languages]
 
-        return sum(ratios) / 2 if self.non_latin_covered_any is True else sum(ratios)
+        return statistics.mean(ratios) / 2 if self.non_latin_covered_any is True else statistics.mean(ratios)
 
     @property
     def coverage(self):
@@ -168,6 +169,14 @@ class ProbeCoherence:
 
                     self.nb_used_occurrence += used_occ
 
+                    # if language == 'Hebrew':
+                    #     logger.debug(
+                    #         'Language {lang} was potentially detected. R1 = {r1} with R2 = {r2}',
+                    #         lang=language,
+                    #         r1=str(round(ratio_unavailable_letters*100, ndigits=3)),
+                    #         r2=str(round((1. - not_respected_rank_coeff) * 100, ndigits=3))
+                    #     )
+
                     for l, o in most_common_cpy:
                         self.covered_letters.add(l)
 
